from alvin_integration.installer.base import AlvinBaseInstaller
from alvin_integration.producers.airflow.config import AirflowProducerConfig


class AlvinAirflowInstaller(AlvinBaseInstaller):
    def __init__(self, *args, **kwargs):
        super(AlvinAirflowInstaller, self).__init__(
            provider_config=AirflowProducerConfig(), *args, **kwargs
        )



# overrides https://github.com/dbt-labs/dbt-core/blob/main/core/dbt/task/run.py#L418
# payload based on https://github.com/OpenLineage/OpenLineage/blob/main/integration/dbt/scripts/dbt-ol
from alvin_integration.helper import log_verbose


def after_run(self, adapter, results):
    log_verbose("start after_run")
    # in on-run-end hooks, provide the value 'database_schemas', which is a
    # list of unique (database, schema) pairs that successfully executed
    # models were in. For backwards compatibility, include the old
    # 'schemas', which did not include database information.
    import uuid
    from datetime import datetime, timezone

    from alvin_integration.adapter import AlvinLineageBackendAdapter
    from dbt.contracts.results import NodeStatus
    from dbt.node_types import RunHookType

    database_schema_set = {
        (r.node.database, r.node.schema)
        for r in results
        if r.node.is_relational
        and r.status not in (NodeStatus.Error, NodeStatus.Fail, NodeStatus.Skipped)
    }

    self._total_executed += len(results)

    extras = {
        "schemas": list({s for _, s in database_schema_set}),
        "results": results,
        "database_schemas": list(database_schema_set),
    }
    with adapter.connection_named("master"):
        self.safe_run_hooks(adapter, RunHookType.End, extras)
    backend_adapter = AlvinLineageBackendAdapter()

    import os

    alvin_platform_id = os.environ.get("ALVIN_PLATFORM_ID")
    alvin_dbt_platform_id = os.environ.get("ALVIN_DBT_PLATFORM_ID")

    log_verbose(f"alvin_platform_id: {alvin_platform_id}")
    log_verbose(f"alvin_dbt_platform_id: {alvin_dbt_platform_id}")

    # Create a single run_id for all results under same dbt run
    run_id = str(uuid.uuid4())
    for result in results:
        payload = {
            "event_time": datetime.now(timezone.utc).isoformat(),
            "event_type": result.status.upper(),
            "producer": "alvin.ai",
            "run": {
                "run_id": run_id,
            },
            "job": {
                "job_id": result.adapter_response.get("job_id"),
                "job_name": f"dbt-run-{result.node.package_name}",
            },
            "alvin_facet": {
                "dbt": {
                    "alvin_platform_id": alvin_dbt_platform_id,
                    "adapter": "bigquery",
                    "job_id": result.adapter_response.get("job_id"),
                    "project_id": result.node.database,
                    "database": result.node.database,
                    "code": result.adapter_response.get("code"),
                    "compiled_sql": result.node.compiled_sql
                    if hasattr(result.node, "compiled_sql")
                    else None,
                    "raw_sql": result.node.raw_sql
                    if hasattr(result.node, "raw_sql")
                    else None,
                    "thread_id": result.thread_id,
                    "execution_time": result.execution_time,
                    "failures": result.failures,
                    "message": result.message,
                    "package_name": result.node.package_name,
                    "root_path": result.node.root_path,
                    "schema": result.node.schema,
                    "fqn": result.node.fqn,
                    "unique_id": result.node.unique_id,
                    "path": result.node.path,
                    "original_file_path": result.node.original_file_path,
                    "name": result.node.name,
                    "alias": result.node.alias,
                    "resource_type": result.node.resource_type,
                },
                "airflow": {
                    "alvin_platform_id": alvin_platform_id,
                },
            },
        }
        log_verbose("start send dbt metadata")

        backend_adapter.send_data(payload, "api/v1/metadata")

        log_verbose("end send dbt metadata")


# overrides https://github.com/dbt-labs/dbt-bigquery/blob/main/dbt/adapters/bigquery/connections.py#L442
def execute(self, sql, auto_begin=False, fetch=None):
    from alvin_integration.producers.dbt.lineage.extractors.bigquery import (
        AlvinBigQueryAdapterResponse,
    )
    from dbt.clients import agate_helper

    sql = self._add_query_comment(sql)
    # auto_begin is ignored on bigquery, and only included for consistency
    query_job, iterator = self.raw_execute(sql, fetch=fetch)

    if fetch:
        table = self.get_table_from_response(iterator)
    else:
        table = agate_helper.empty_table()

    message = "OK"
    code = None
    num_rows = None
    bytes_processed = None

    if query_job.statement_type == "CREATE_VIEW":
        code = "CREATE VIEW"

    elif query_job.statement_type == "CREATE_TABLE_AS_SELECT":
        conn = self.get_thread_connection()
        client = conn.handle
        query_table = client.get_table(query_job.destination)
        code = "CREATE TABLE"
        num_rows = query_table.num_rows
        num_rows_formated = self.format_rows_number(num_rows)
        bytes_processed = query_job.total_bytes_processed
        processed_bytes = self.format_bytes(bytes_processed)
        message = f"{code} ({num_rows_formated} rows, {processed_bytes} processed)"

    elif query_job.statement_type == "SCRIPT":
        code = "SCRIPT"
        bytes_processed = query_job.total_bytes_processed
        message = f"{code} ({self.format_bytes(bytes_processed)} processed)"

    elif query_job.statement_type in ["INSERT", "DELETE", "MERGE", "UPDATE"]:
        code = query_job.statement_type
        num_rows = query_job.num_dml_affected_rows
        num_rows_formated = self.format_rows_number(num_rows)
        bytes_processed = query_job.total_bytes_processed
        processed_bytes = self.format_bytes(bytes_processed)
        message = f"{code} ({num_rows_formated} rows, {processed_bytes} processed)"

    elif query_job.statement_type == "SELECT":
        conn = self.get_thread_connection()
        client = conn.handle
        # use anonymous table for num_rows
        query_table = client.get_table(query_job.destination)
        code = "SELECT"
        num_rows = query_table.num_rows
        num_rows_formated = self.format_rows_number(num_rows)
        bytes_processed = query_job.total_bytes_processed
        processed_bytes = self.format_bytes(bytes_processed)
        message = f"{code} ({num_rows_formated} rows, {processed_bytes} processed)"

    response = AlvinBigQueryAdapterResponse(
        _message=message,
        rows_affected=num_rows,
        code=code,
        bytes_processed=bytes_processed,
        job_id=query_job.job_id if hasattr(query_job, "job_id") else None,
    )

    return response, table



import airflow


def _run_finished_callback(self, error=None):
    """
    Alvin patched version of _run_finished_callback from TaskInstance.

    This method will inject the call
    to alvin_callback in case the task has failed.
    """

    import logging
    import traceback

    from airflow.utils.state import State

    from alvin_integration.helper import AlvinLoggerAdapter
    from alvin_integration.producers.airflow.lineage.backend import alvin_callback

    log = AlvinLoggerAdapter(logging.getLogger(__name__), {})

    context = None

    try:
        # call unpatched function
        # see for details
        # https://airflow.apache.org/docs/apache-airflow/2.1.0/_api/airflow/models/taskinstance/index.html#airflow.models.taskinstance.TaskInstance._run_finished_callback
        if self.state == State.FAILED:
            task = self.task
            if task.on_failure_callback is not None:
                context = self.get_template_context()
                context["exception"] = error
                task.on_failure_callback(context)
        elif self.state == State.SUCCESS:
            task = self.task
            if task.on_success_callback is not None:
                context = self.get_template_context()
                task.on_success_callback(context)
        elif self.state == State.UP_FOR_RETRY:
            task = self.task
            if task.on_retry_callback is not None:
                context = self.get_template_context()
                context["exception"] = error
                task.on_retry_callback(context)
    except Exception:
        log.error(f"Default Callback Error: {traceback.format_exc()}")

    if not context:
        context = self.get_template_context()

    alvin_callback(context=context, operator=self.task)

def get_all_dagmodels(self, session=None):
    from airflow.models.dag import DagModel
    from airflow.utils.session import provide_session
    

    @provide_session
    def get_session(session):
        return session

    if not session:
        session = get_session()

    return session.query(DagModel).all()

def handle_failure(
    self,
    error,
    test_mode=None,
    context=None,
    force_fail=False,
    session=None,  # noqa
):
    """
    Alvin patched version of handle_failure from TaskInstance.

    This method will inject the call
    to alvin_callback in case the task has failed.
    """
    import logging

    from alvin_integration.helper import AlvinLoggerAdapter
    from alvin_integration.producers.airflow.lineage.backend import alvin_callback

    log = AlvinLoggerAdapter(logging.getLogger(__name__), {})

    @airflow.utils.db.provide_session
    def get_session(session):
        return session

    if not session:
        session = get_session()
    from airflow.models.log import Log
    from airflow.models.taskfail import TaskFail
    from airflow.settings import Stats
    from airflow.utils import timezone
    from airflow.utils.state import State

    if test_mode is None:
        test_mode = self.test_mode
    if context is None:
        context = self.get_template_context()

    log.exception(error)
    task = self.task
    self.end_date = timezone.utcnow()
    self.set_duration()
    Stats.incr("operator_failures_{}".format(task.__class__.__name__), 1, 1)
    Stats.incr("ti_failures")
    if not test_mode:
        session.add(Log(State.FAILED, self))

    # Log failure duration
    session.add(TaskFail(task, self.execution_date, self.start_date, self.end_date))

    if context is not None:
        context["exception"] = error

    # Set state correctly and figure out how to log it,
    # what callback to call if any, and how to decide whether to email

    # Since this function is called only when the TaskInstance state is running,
    # try_number contains the current try_number (not the next). We
    # only mark task instance as FAILED if the next task instance
    # try_number exceeds the max_tries ... or if force_fail is truthy

    if force_fail or not self.is_eligible_to_retry():
        self.state = State.FAILED
        if force_fail:
            log_message = "Immediate failure requested. Marking task as FAILED."
        else:
            log_message = "Marking task as FAILED."
        email_for_state = task.email_on_failure
        callback = task.on_failure_callback
    else:
        self.state = State.UP_FOR_RETRY
        log_message = "Marking task as UP_FOR_RETRY."
        email_for_state = task.email_on_retry
        callback = task.on_retry_callback

    log.info(
        "%s dag_id=%s, task_id=%s, execution_date=%s, start_date=%s, end_date=%s",
        log_message,
        self.dag_id,
        self.task_id,
        self._safe_date("execution_date", "%Y%m%dT%H%M%S"),
        self._safe_date("start_date", "%Y%m%dT%H%M%S"),
        self._safe_date("end_date", "%Y%m%dT%H%M%S"),
    )
    if email_for_state and task.email:
        try:
            self.email_alert(error)
        except Exception as e2:
            log.error("Failed to send email to: %s", task.email)
            log.exception(e2)

    # Handling callbacks pessimistically
    if callback:
        try:
            callback(context)
        except Exception as e3:
            log.error("Failed at executing callback")
            log.exception(e3)

    if not test_mode:
        session.merge(self)
    session.commit()

    alvin_callback(context=context, operator=self.task, is_airflow_legacy=True)  # noqa


def handle_callback(self, dagrun, success=True, reason=None, session=None):
    """
    Alvin patched version of handle_callback from DAG.

    This method will inject the call
    to alvin_callback in case the task has failed.
    """
    import logging

    from alvin_integration.helper import AlvinLoggerAdapter
    from alvin_integration.producers.airflow.lineage.backend import alvin_callback

    log = AlvinLoggerAdapter(logging.getLogger(__name__), {})

    task_instance_list = dagrun.get_task_instances()

    if not success:
        is_airlfow_legacy = True

        for ti in task_instance_list:
            ti.task = self.get_task(ti.task_id)
            context = ti.get_template_context(session=session)
            alvin_callback(
                context=context,
                operator=ti.task,
                is_airflow_legacy=is_airlfow_legacy,  # noqa
            )

    callback = self.on_success_callback if success else self.on_failure_callback  # noqa

    if callback:
        log.info("Executing dag callback function: {}".format(callback))
        task_instance_list = dagrun.get_task_instances()
        ti = task_instance_list[-1]  # get first TaskInstance of DagRun
        ti.task = self.get_task(ti.task_id)
        context = ti.get_template_context(session=session)
        context.update({"reason": reason})
        callback(context)


def pre_execute(self, context):
    """
    Alvin patched version of pre_execute from BigQueryOperator.

    This method will create the hook in the operator object
    and instantiate the bq_cursor.
    """
    from airflow.contrib.hooks.bigquery_hook import BigQueryHook

    @airflow.lineage.prepare_lineage
    def prepare_lineage(session):
        pass

    prepare_lineage()
    self.hook = BigQueryHook(
        bigquery_conn_id=self.bigquery_conn_id,
        use_legacy_sql=self.use_legacy_sql,
        delegate_to=self.delegate_to,
        location=self.location,
    )
    conn = self.hook.get_conn()
    self.bq_cursor = conn.cursor()


def get_client(self, project_id=None, location=None):
    """
    Alvin patched to add the method get_client to the BigQueryOperator.

    This method will create crate a Google Client using the
    reusing the Airflow Google connection.
    """
    from google.cloud.bigquery import Client

    return Client(
        project=project_id,
        location=location,
        credentials=self._get_credentials(),
    )


def run(self, sql, autocommit=False, parameters=None):
    """
    Runs a command or a list of commands. Pass a list of sql
    statements to the sql parameter to get them to execute
    sequentially
    :param sql: the sql statement to be executed (str) or a list of
        sql statements to execute
    :type sql: str or list
    :param autocommit: What to set the connection's autocommit setting to
        before executing the query.
    :type autocommit: bool
    :param parameters: The parameters to render the SQL query with.
    :type parameters: mapping or iterable
    """
    import logging
    import sys
    from contextlib import closing

    from past.builtins import basestring

    from alvin_integration.helper import AlvinLoggerAdapter

    log = AlvinLoggerAdapter(logging.getLogger(__name__), {})

    self.query_ids = []

    if isinstance(sql, basestring):
        sql = [sql]

    with closing(self.get_conn()) as conn:
        if self.supports_autocommit:
            self.set_autocommit(conn, autocommit)

        with closing(conn.cursor()) as cur:
            for s in sql:
                if sys.version_info[0] < 3:
                    s = s.encode("utf-8")
                if parameters is not None:
                    log.info("{} with parameters {}".format(s, parameters))
                    cur.execute(s, parameters)
                else:
                    log.info(s)
                    cur.execute(s)
                query_id = cur.sfqid
                log.info("Rows affected: %s", cur.rowcount)
                log.info("Snowflake query id: %s", query_id)
                self.query_ids.append(query_id)

        # If autocommit was set to False for db that supports autocommit,
        # or if db does not supports autocommit, we do a manual commit.
        if not self.get_autocommit(conn):
            conn.commit()


def get_hook(self):
    from airflow.contrib.hooks.snowflake_hook import SnowflakeHook

    if not hasattr(self, "hook"):
        self.hook = SnowflakeHook(
            snowflake_conn_id=self.snowflake_conn_id,
            warehouse=self.warehouse,
            database=self.database,
            role=self.role,
            schema=self.schema,
            authenticator=self.authenticator,
        )
    return self.hook


def _add_auth(self, api_key: str):
    if "alvin" in api_key:
        alvin_api_key = api_key.split("-alvin-")
        self.session.headers.update({"X-API-KEY": f"{alvin_api_key[1]}"})
    else:
        self.session.headers.update({"Authorization": f"Bearer {api_key}"})


def get_or_create_openlineage_client(self):
    import logging
    import os

    from openlineage.client import OpenLineageClient, OpenLineageClientOptions

    from alvin_integration.helper import AlvinLoggerAdapter

    log = AlvinLoggerAdapter(logging.getLogger(__name__), {})

    if not self._client:
        alvin_url = os.getenv("ALVIN_URL")
        alvin_api_key = os.getenv("ALVIN_API_KEY")
        if alvin_url:
            log.info(f"Sending lineage events to Alvin Backend: {alvin_url}")
            self._client = OpenLineageClient(
                alvin_url,
                OpenLineageClientOptions(api_key=f"key-alvin-{alvin_api_key}"),
            )
            return self._client
        # Backcomp with Marquez integration
        marquez_url = os.getenv("MARQUEZ_URL")
        marquez_api_key = os.getenv("MARQUEZ_API_KEY")

        if marquez_url:
            log.info(f"Sending lineage events to {marquez_url}")
            self._client = OpenLineageClient(
                marquez_url, OpenLineageClientOptions(api_key=marquez_api_key)
            )
        else:
            self._client = OpenLineageClient.from_environment()
    return self._client


def update_state(self, session=None, execute_callbacks=True):
    """
    Determines the overall state of the DagRun based on the state
    of its TaskInstances.
    :param session: Sqlalchemy ORM Session
    :type session: Session
    :param execute_callbacks: Should dag callbacks (success/failure, SLA etc) be invoked
        directly (default: true) or recorded as a pending request in the ``callback`` property
    :type execute_callbacks: bool
    :return: Tuple containing tis that can be scheduled in the current loop & `callback` that
        needs to be executed
    """
    import logging
    from typing import Optional

    from airflow.stats import Stats
    from airflow.utils import callback_requests, timezone
    from airflow.utils.state import DagRunState, State, TaskInstanceState

    from alvin_integration.helper import AlvinLoggerAdapter
    from alvin_integration.producers.airflow.lineage.backend import (
        alvin_dag_run_extractor,
    )

    log = AlvinLoggerAdapter(logging.getLogger(__name__), {})

    # Callback to execute in case of Task Failures
    callback: Optional[callback_requests.DagCallbackRequest] = None

    @airflow.utils.session.provide_session
    def get_session(session):
        return session

    if not session:
        session = get_session()

    start_dttm = timezone.utcnow()
    self.last_scheduling_decision = start_dttm
    with Stats.timer(f"dagrun.dependency-check.{self.dag_id}"):
        dag = self.get_dag()
        info = self.task_instance_scheduling_decisions(session)

        tis = info.tis
        schedulable_tis = info.schedulable_tis
        changed_tis = info.changed_tis
        finished_tasks = info.finished_tasks
        unfinished_tasks = info.unfinished_tasks

        none_depends_on_past = all(not t.task.depends_on_past for t in unfinished_tasks)
        none_task_concurrency = all(
            t.task.max_active_tis_per_dag is None for t in unfinished_tasks
        )
        none_deferred = all(t.state != State.DEFERRED for t in unfinished_tasks)

        if (
            unfinished_tasks
            and none_depends_on_past
            and none_task_concurrency
            and none_deferred
        ):
            # small speed up
            are_runnable_tasks = (
                schedulable_tis
                or self._are_premature_tis(unfinished_tasks, finished_tasks, session)
                or changed_tis
            )

    leaf_task_ids = {t.task_id for t in dag.leaves}
    leaf_tis = [ti for ti in tis if ti.task_id in leaf_task_ids]

    # if all roots finished and at least one failed, the run failed
    if not unfinished_tasks and any(
        leaf_ti.state in State.failed_states for leaf_ti in leaf_tis
    ):
        log.error("Marking run %s failed", self)
        self.set_state(State.FAILED)
        if execute_callbacks:
            dag.handle_callback(
                self, success=False, reason="task_failure", session=session
            )
        elif dag.has_on_failure_callback:
            callback = callback_requests.DagCallbackRequest(
                full_filepath=dag.fileloc,
                dag_id=self.dag_id,
                run_id=self.run_id,
                is_failure_callback=True,
                msg="task_failure",
            )

    # if all leaves succeeded and no unfinished tasks, the run succeeded
    elif not unfinished_tasks and all(
        leaf_ti.state in State.success_states for leaf_ti in leaf_tis
    ):
        log.info("Marking run %s successful", self)
        self.set_state(State.SUCCESS)
        if execute_callbacks:
            dag.handle_callback(self, success=True, reason="success", session=session)
        elif dag.has_on_success_callback:
            callback = callback_requests.DagCallbackRequest(
                full_filepath=dag.fileloc,
                dag_id=self.dag_id,
                run_id=self.run_id,
                is_failure_callback=False,
                msg="success",
            )

    # if *all tasks* are deadlocked, the run failed
    elif (
        unfinished_tasks
        and none_depends_on_past
        and none_task_concurrency
        and none_deferred
        and not are_runnable_tasks
    ):
        log.error("Deadlock; marking run %s failed", self)
        self.set_state(State.FAILED)
        if execute_callbacks:
            dag.handle_callback(
                self, success=False, reason="all_tasks_deadlocked", session=session
            )
        elif dag.has_on_failure_callback:
            callback = callback_requests.DagCallbackRequest(
                full_filepath=dag.fileloc,
                dag_id=self.dag_id,
                run_id=self.run_id,
                is_failure_callback=True,
                msg="all_tasks_deadlocked",
            )

    # finally, if the roots aren't done, the dag is still running
    else:
        self.set_state(State.RUNNING)

    if self._state == State.FAILED or self._state == State.SUCCESS:
        msg = (
            "Alvin DagRun Finished: dag_id=%s, execution_date=%s, run_id=%s, "
            "run_start_date=%s, run_end_date=%s, run_duration=%s, "
            "state=%s, external_trigger=%s, run_type=%s, "
            "data_interval_start=%s, data_interval_end=%s, dag_hash=%s"
        )
        log.info(
            msg,
            self.dag_id,
            self.execution_date,
            self.run_id,
            self.start_date,
            self.end_date,
            (self.end_date - self.start_date).total_seconds()
            if self.start_date and self.end_date
            else None,
            self._state,
            self.external_trigger,
            self.run_type,
            self.data_interval_start,
            self.data_interval_end,
            self.dag_hash,
        )

    alvin_dag_run_extractor(dag_run=self)
    self._emit_true_scheduling_delay_stats_for_finished_state(finished_tasks)
    self._emit_duration_stats_for_finished_state()

    session.merge(self)

    return schedulable_tis, callback



"""
This module contains the data models that will be produced
and sent to the Alvin Backend.
"""
import importlib
import os
from dataclasses import dataclass
from typing import List


@dataclass
class AlvinAirflowTaskExecution:
    task_id: str
    dag_id: str
    run_id: str
    start_date: str
    end_date: str
    duration: float
    try_number: int
    pool: str
    operator: str
    state: str
    run_id: str
    platform_id: str
    max_tries: int
    alvin_package_version: str


@dataclass
class AlvinBaseLineageDetails:
    execution: AlvinAirflowTaskExecution
    connection_id: str


@dataclass
class AlvinBigQueryLineageDetails(AlvinBaseLineageDetails):
    job_id: str
    project_id: str


@dataclass
class AlvinSnowflakeLineageDetails(AlvinBaseLineageDetails):
    query_ids: List[str]


@dataclass
class AlvinFacet:
    alvin: AlvinBaseLineageDetails


@dataclass
class AlvinPatch:
    package_name: str
    function: any
    supported_versions: List[str]
    destination_path: str
    is_required: bool = False

    @property
    def destination(self):
        module_name, class_name = self.destination_path.rsplit(".", 1)
        return getattr(importlib.import_module(module_name), class_name)

    def __repr__(self):
        return (
            f"package_name: {self.package_name}, "
            f"patch: {self.function}, "
            f"supported_versions: {self.supported_versions}, "
            f"destination: {self.destination}"
        )


@dataclass
class AlvinLineageConfig:
    package_name: str
    env_name: str
    env_value: str
    supported_versions: List[str]

    def set_lineage(self):
        os.environ[self.env_name] = self.env_value

    def __repr__(self):
        return (
            f"package_name: {self.package_name}, "
            f"env_name: {self.env_name}, "
            f"env_value: {self.env_value}, "
            f"supported_versions: {self.supported_versions}, "
        )


@dataclass
class AlvinPipelineCreator:
    package_name: str
    function: any
    supported_versions: List[str]

    def __repr__(self):
        return (
            f"function: {self.function}, "
            f"supported_versions: {self.supported_versions}"
        )



from dataclasses import dataclass

from dbt.adapters.bigquery.connections import BigQueryAdapterResponse


@dataclass
class AlvinBigQueryAdapterResponse(BigQueryAdapterResponse):
    job_id: str = None
    alvin_platform_id: str = None



import os
from typing import List

import requests
from alvin_integration.helper import log_verbose
from sqlalchemy.orm.session import Session

from alvin_integration.producers.airflow.config import (
    ALVIN_BACKEND_API_KEY,
    ALVIN_BACKEND_API_URL,
)

ALVIN_PLATFORM_ID = os.getenv("ALVIN_PLATFORM_ID")


def get_airflow_version():
    import airflow

    return airflow.__version__


def get_dag_source_code(dag_location):
    if dag_location:
        with open(dag_location, "r") as f:
            content = f.read()
        return content


def extract_dag_metadata(session: Session):
    from airflow.models.dag import DagModel
    from airflow.serialization.serialized_objects import SerializedDAG

    alvin_backend_metadata_url = f"{ALVIN_BACKEND_API_URL}/api/v1/metadata"
    entities = []
    all_dags_models: List[DagModel] = DagModel().get_all_dagmodels(session)

    for dag_model in all_dags_models:
        dag = dag_model.serialized_dag.dag
        dag_metadata = SerializedDAG.to_dict(dag)

        enrich_dag_metadata(dag_metadata, dag_model, dag)
        entities.append(dag_metadata["dag"])
    payload = {
        "alvin_platform_id": ALVIN_PLATFORM_ID,
        "facets": {
            "airflow_version": get_airflow_version(),
            "platform_type": "AIRFLOW",
        },
        "entities": entities,
    }
    requests.post(
        alvin_backend_metadata_url,
        json=payload,
        headers={"X-API-KEY": ALVIN_BACKEND_API_KEY},
    )
    log_verbose(f"Metadata extracted: {payload}")

def enrich_dag_metadata(dag_metadata, dag_model, dag):
    "Adds metadata fields to dag metadata"

    dag_metadata["dag"]["source_code"] = get_dag_source_code(
            dag_metadata["dag"].get("fileloc")
        )
    dag_metadata["dag"]["is_active"] = dag_model.is_active
    dag_metadata["dag"]["leaves"] = [task.task_id for task in dag.leaves]
    dag_metadata["dag"]["roots"] = [task.task_id for task in dag.roots]
    for task in dag_metadata["dag"].get("tasks", []):
        task_instance = dag.task_dict.get(task["task_id"])
        task["_upstream_task_ids"] = list(task_instance.upstream_task_ids) if task_instance else []
        task["operator_type"] = task_instance.task_type if task_instance else task.get("_task_type")



from dbt.adapters.base import AdapterPlugin
from dbt.adapters.bigquery.column import BigQueryColumn  # noqa
from dbt.adapters.bigquery.connections import BigQueryConnectionManager  # noqa
from dbt.adapters.bigquery.connections import BigQueryCredentials
from dbt.adapters.bigquery.impl import BigQueryAdapter, GrantTarget  # noqa
from dbt.adapters.bigquery.relation import BigQueryRelation  # noqa
from dbt.include import bigquery

from alvin_integration.producers.dbt.lineage.backend import AlvinBigQueryAdapter

Plugin = AdapterPlugin(
    adapter=AlvinBigQueryAdapter,
    credentials=BigQueryCredentials,
    include_path=bigquery.PACKAGE_PATH,
)



import logging

from alvin_integration.helper import log_verbose
from openlineage.airflow.extractors.base import TaskMetadata
from openlineage.airflow.extractors.snowflake_extractor import SnowflakeExtractor
from openlineage.airflow.utils import (
    get_connection,
    get_normalized_postgres_connection_uri,
    safe_import_airflow,
)
from openlineage.client.facet import SqlJobFacet
from openlineage.common.dataset import Dataset, Source
from openlineage.common.sql import SqlMeta, SqlParser

from alvin_integration.models import AlvinFacet, AlvinSnowflakeLineageDetails
from alvin_integration.producers.airflow.lineage.extractors.mixins import (
    AlvinAirflowExtractorMixin,
)

logger = logging.getLogger(__name__)


class AlvinSnowflakeBaseExtractor(AlvinAirflowExtractorMixin, SnowflakeExtractor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _get_hook(self):
        raise NotImplementedError

    def get_connection_id(self):
        return self.operator.snowflake_conn_id if self.operator else None

    def build_facet(self, task_instance):
        return AlvinFacet(
            alvin=AlvinSnowflakeLineageDetails(
                query_ids=self._get_query_ids(),
                execution=self.get_execution_details(task_instance),
                connection_id=self.get_connection_id(),
            )
        )

    def _get_database(self) -> str:
        return self._get_hook().database

    def _get_authority(self) -> str:
        return self._get_hook().snowflake_conn_id


class AlvinSnowflakeExtractor(AlvinSnowflakeBaseExtractor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _get_hook(self):
        return self.operator.get_db_hook()

    def _get_query_ids(self):
        return self.operator.query_ids

    def get_airflow_run_id(self, task_instance):
        return task_instance.get_dagrun().run_id


class AlvinSnowflakeLegacyExtractor(AlvinSnowflakeBaseExtractor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        log_verbose("AlvinSnowflakeLegacyExtractor")

    def _get_hook(self):
        return self.operator.get_hook()

    def extract(self) -> TaskMetadata:

        sql_meta: SqlMeta = SqlParser.parse(self.operator.sql, self.default_schema)

        self.conn = get_connection(self._conn_id())

        source = Source(
            scheme=self._get_scheme(),
            authority=self._get_authority(),
            connection_url=self._get_connection_uri(),
        )

        database = self.operator.database
        if not database:
            database = self._get_database()

        inputs = [
            Dataset.from_table(
                source=source,
                table_name=in_table_schema.table_name.name,
                schema_name=in_table_schema.schema_name,
                database_name=database,
            )
            for in_table_schema in self._get_table_schemas(sql_meta.in_tables)
        ]
        outputs = [
            Dataset.from_table_schema(
                source=source, table_schema=out_table_schema, database_name=database
            )
            for out_table_schema in self._get_table_schemas(sql_meta.out_tables)
        ]
        query_ids = self._get_query_ids()

        return TaskMetadata(
            name=f"{self.operator.dag_id}.{self.operator.task_id}",
            inputs=[ds.to_openlineage_dataset() for ds in inputs],
            outputs=[ds.to_openlineage_dataset() for ds in outputs],
            job_facets={
                "sql": SqlJobFacet(self.operator.sql),
                "queries_ids": query_ids if query_ids else [],
            },
        )

    def _get_query_ids(self):
        hook = self.operator.get_hook()
        log_verbose(hook)
        if hasattr(hook, "query_ids"):
            log_verbose(f"Query ids found: {hook.query_ids}")
            return hook.query_ids
        else:
            log_verbose(f"Query ids not found")
            return []

    def get_airflow_run_id(self, task_instance):
        return task_instance.get_dagrun().run_id



from alvin_integration.interfaces.config import AbstractProducerConfig
from alvin_integration.models import AlvinPatch
from alvin_integration.producers.dbt.patch.functions import after_run, execute
from alvin_integration.producers.dbt.patch.functions_legacy import execute as execute_legacy


class DBTProducerConfig(AbstractProducerConfig):
    @property
    def producer_name(self):
        return "dbt"

    def get_patching_list(self):
        return [
            AlvinPatch(
                package_name="dbt-core",
                function=after_run,
                supported_versions=["1.0.0", "1.0.1", "1.2.0", "1.2.1", "1.2.2"],
                destination_path="dbt.task.run.RunTask",
            ),
            AlvinPatch(
                package_name="dbt-core",
                function=execute,
                supported_versions=["1.2.0", "1.2.1", "1.2.2"],
                destination_path="dbt.adapters.bigquery.connections.BigQueryConnectionManager",
            ),
            AlvinPatch(
                package_name="dbt-core",
                function=execute_legacy,
                supported_versions=["1.0.0", "1.0.1"],
                destination_path="dbt.adapters.bigquery.connections.BigQueryConnectionManager",
            ),
        ]

    def get_lineage_config(self):
        pass

    def get_target_packages(self):
        return ["dbt-core"]

    def get_target_pipelines(self):
        pass
